@@ -2,6 +2,8 @@ import types
 from Mapbox import vector_tile_pb2
 from shapely.wkb import loads as load_wkb
 from shapely.wkt import loads as load_wkt
+from shapely.geometry.base import BaseGeometry
+from numbers import Number
 
 from math import floor, fabs
 from array import array
@@ -28,42 +30,73 @@ class VectorTile:
         self.layer.name    = layer_name
         self.layer.version = 2
         self.layer.extent  = self.extents
-        self.feature_count = 0
         self.keys   = []
         self.values = []
 
         for feature in features:
-            self.addFeature(feature)
 
-    def addFeature(self, feature):
+            # skip missing or empty geometries
+            geometry_spec = feature.get('geometry')
+            if geometry_spec is None:
+                continue
+            shape = self._load_geometry(geometry_spec)
+            if shape is None:
+                raise NotImplementedError(
+                    'Can\'t do geometries that are not wkt, wkb, or shapely '
+                    'geometries')
+            if shape.is_empty:
+                continue
+            # if we are a multipolygon, we will create separate
+            # features for each polygon and duplicate properties
+            if shape.type == 'MultiPolygon':
+                exploded_features = self.explode_multipolygon_features(
+                    feature, shape)
+                for exploded_feature in exploded_features:
+                    self.addFeature(exploded_feature,
+                                    exploded_feature['geometry'])
+            else:
+                self.addFeature(feature, shape)
+
+    def explode_multipolygon_features(self, feature, shape):
+        assert shape.type == 'MultiPolygon'
+        exploded_features = []
+        properties = feature['properties']
+        feature_id = feature.get('id')
+        for geometry in shape.geoms:
+            new_feature = dict(properties=properties, geometry=geometry)
+            if feature_id is not None:
+                new_feature['id'] = feature_id
+            exploded_features.append(new_feature)
+        return exploded_features
+
+    def _load_geometry(self, geometry_spec):
+        if isinstance(geometry_spec, BaseGeometry):
+            return geometry_spec
+
+        try:
+            return load_wkb(geometry_spec)
+        except:
+            try:
+                return load_wkt(geometry_spec)
+            except:
+                return None
+
+    def addFeature(self, feature, shape):
         f = self.layer.features.add()
-        self.feature_count += 1
-        f.id = self.feature_count
-        
-        # osm_id or the hash can be passed in as a feature['id']
-        if feature.has_key("id"):
-            feature[1].update(uid=feature["id"])
+
+        fid = feature.get('id')
+        if fid is not None:
+            if isinstance(fid, Number) and fid >= 0:
+                f.id = fid
 
         # properties
-        if feature.has_key("properties"):
-            self._handle_attr(self.layer, f, feature["properties"])
+        properties = feature.get('properties')
+        if properties is not None:
+            self._handle_attr(self.layer, f, properties)
 
-        # geometry
-        if (feature.has_key("geometry")):
-            shape = {}
-            for load in [load_wkb, load_wkt]:
-                try:
-                    shape = load(feature["geometry"])
-                    break
-                except:
-                    pass
+        f.type = self._get_feature_type(shape)
+        self._geo_encode(f, shape)
 
-            if shape:
-                f.type = self._get_feature_type(shape)
-                self._geo_encode(f, shape)
-            else: 
-                raise NotImplementedError("Can't do geometries that are not wkt or wkb")
-                
     def _get_feature_type(self, shape):
         if shape.type == 'Point' or shape.type == 'MultiPoint':
             return self.tile.Point
@@ -90,9 +123,12 @@ class VectorTile:
                     if (isinstance(v,bool)):
                         val = layer.values.add()
                         val.bool_value = v
-                    elif (isinstance(v,str)) or (isinstance(v,unicode)):
+                    elif (isinstance(v,str)):
                         val = layer.values.add()
                         val.string_value = unicode(v,'utf8')
+                    elif (isinstance(v,unicode)):
+                        val = layer.values.add()
+                        val.string_value = v
                     elif (isinstance(v,int)) or (isinstance(v,long)):
                         val = layer.values.add()
                         val.int_value = v
